-- vim: ts=4 sw=4

Set = {}

function Set.new (t)
	local set = {}
	setmetatable(set, Set)
	for _,v in ipairs(t) do
		Set.add(set, v)
	end
	return set
end

Set.copy = Set.new
Set.getn = table.getn

function Set.add (s,v)
	if not s[v] then
		table.insert(s, v)
		s[v] = Set.getn(s)
	end
	return set
end

function Set.remove (s,v)
	if s[v] then
		table.remove(s, s[v])
		s[v] = nil
	end
	return set
end

function Set.weaken (s,v) -- semi-remove the element
	s[v] = nil
end

function Set.fixup(s) -- purge removed elements
	local removed
	repeat
		removed = 0
		for i,v in ipairs(s) do
			if not s[v] then
				table.remove(s, i)
				removed = 1
				break
			end
		end
	until removed == 0
end

function Set.each (s)
	local i = 0
	local n = Set.getn(s)
	return function ()
		i = i + 1
		if i <= n then 
			return s[i]
		end
	end
end

function Set.union (a,b)
	local set = Set.new{}
	if type(b) ~= "table" then b = Set.new{b} end
	for v in Set.each(a) do Set.add(set, v) end
	for v in Set.each(b) do Set.add(set, v) end
	return set
end

function Set.intersection (a,b)
	local set = Set.new{}
	for v in Set.each(a) do
		if b[v] then
			Set.add(set, v)
		end
	end
	return set
end

Set.__add = Set.union
Set.__mul = Set.intersection

function Set.map (s,f)
	local set = Set.new{}
	for v in Set.each(s) do
		set = set + f(v)
	end
	return set
end

function Set.sort (s,f)
	table.sort(s, f)
	for i,v in ipairs(s) do
		s[v] = i
	end
end

function Set.tostring (s)
	local str = "{ "
	local sep = ""
	for v in Set.each(s) do
		str = str .. sep .. v
		sep = ", "
	end
	return str .. " }"
end

Set.__tostring = Set.tostring

function Set.print (s)
	print(Set.tostring(s))
end

return Set
