#!/usr/bin/lua5
-- vim: ts=4 sw=4

require "RPM"
db = RPM.database()

function verbose(...)
	io.stderr:write(table.concat(arg), "\n")
end

require "Set"

function expand(packages)
	local expanded = Set.copy(packages)
	local pass = 0
	repeat
		packages = Set.copy(expanded)
		local npkg = Set.getn(packages)
		pass = pass + 1
		verbose("\tExpanding (pass ", pass, ", ", npkg, " packages)...")
		for pkg in Set.each(packages) do
			local rpm = db:pkg(pkg)
			local req = rpm:requires()
			req = Set.new(req)
			for dep in Set.each(req) do
				while true do -- try to break early
					if expanded[dep] then break; end
					if db:pkg(dep) then -- real package
						verbose(pkg, " -> ", dep)
						expanded = expanded + dep
					else -- virtual package
						local prov = db:whatprovides(dep)
						prov = Set.new(prov)
						prov = Set.map(prov, function (hdr) return hdr:tag("name") end)
						if Set.getn(prov * expanded) > 0 then
							break
						end
						if Set.getn(prov) > 1 then
							verbose("Warning: ", dep, " provided by ", Set.tostring(prov))
						elseif Set.getn(prov) == 1 then
							verbose(pkg, " -> ", dep, " -> ", prov[1])
							expanded = expanded + prov
						end
					end
					break
				end
			end
		end
	until Set.getn(packages) == Set.getn(expanded)
	return expanded
end

function qR(pkg) return db:pkg(pkg):requires() end
function qP(pkg) return db:pkg(pkg):provides() end

function squeeze (packages)
	Set.sort(packages, function (a,b)
			local na = table.getn(qR(a))
			local nb = table.getn(qR(b))
			return na < nb
		end )
	for p0 in Set.each(packages) do -- prey
		verbose("\tSqueezing ", p0, " ...")
		for pN in Set.each(packages) do -- raptor
			if not packages[p0] then break end
			if p0 ~= pN then
				local req0, prov0 = Set.new(qR(p0)), Set.new(qP(p0))
				local reqN, provN = Set.new(qR(pN)), Set.new(qP(pN))
				local consume = Set.new{p0} * reqN
				local consumex = prov0 * reqN
				local strive = Set.new{pN} * req0
				local strivex = provN * req0
				local function has (s) return Set.getn(s) > 0 end
				if (has(consume) and (not (has(strive) or has(strivex)) or packages[pN])) then
					verbose(p0, " < ", pN)
					Set.weaken(packages, p0)
					break
				end
				if (has(consumex) and (not (has(strive) or has(strivex)) or packages[pN])) then
					verbose(p0, " < ", consumex[1], " < ", pN)
					Set.weaken(packages, p0)
					break
				end
			end
		end
	end
	Set.fixup(packages)
	return packages
end

local packages = Set.new(arg)
local expanded = expand(packages)
print(expanded)
local sq = squeeze(expanded)
print(sq)
